from __future__ import annotations

from pathlib import Path
from typing import Union, Optional, Tuple, Any

import matplotlib.pyplot as plt    # type: ignore

from . import dataset as ds
from . import pgplot
from .pgplot import (get_pha, get_properties,
                     style_axes)


# -- READ -----------------------------------------------

def read(path: Union[str, Path],
         expno: int,
         procno: int) -> ds.TDatasetnD:
    """
    Creates a Dataset object from the spectrum name folder, expno, and procno.
    """
    p = Path(path) / str(expno) / "pdata" / str(procno)
    return read_abs(p)


def read_abs(path: Union[str, Path]
             ) -> ds.TDatasetnD:
    """
    Creates a Dataset object from the spectrum procno folder.
    """
    p = Path(path)
    # Figure out which type of spectrum it is.
    if not (p / "procs").exists() or not (p.parents[1] / "acqus").exists():
        raise ValueError(f"Invalid path to spectrum {p}: procs or acqus not found")
    if (p.parents[1] / "ser").exists() and (p / "2rr").exists():
        return ds.Dataset2D(p)
    elif (p / "1r").exists() and (p / "used_from").exists():
        try:
            return ds.Dataset1DProj(p)
        # For processed pure shift data, Dataset1DProj will raise a ValueError
        # as it cannot find the projection dimension in the 'used_from' file.
        # The structure of acqus and procs files are more similar to a standard
        # 1D spectrum, so we can simply use Dataset1D instead.
        except ValueError:
            return ds.Dataset1D(p)
    elif (p.parents[1] / "fid").exists() and (p / "1r").exists():
        return ds.Dataset1D(p)
    else:
        raise ValueError(f"Invalid path to spectrum {p}: data files not found")


# -- PLOT ----------------------------------------------

def mkplot(ax: Any = None,
           figsize: Optional[Tuple[float, float]] = None,
           close: bool = True,
           empty_pha: bool = True,
           **kwargs):
    """
    Delegates to _mkplot1d() or _mkplot2d() as necessary.
    """
    # Close open figures, *unless* an axes was given, in which case
    # we assume that the user isn't keen on having that closed!
    # Useful e.g. when doing subplots.
    if close and ax is None:
        plt.close("all")
    # Reset plot properties
    pgplot._reset_properties()

    PHA = get_pha()
    if len(PHA.plot_queue) == 0:
        raise ValueError("No spectra have been staged yet.")
    else:
        if close and figsize is not None and ax is None:
            plt.figure(figsize=figsize)
        if ax is None:
            ax = plt.gca()
        if isinstance(PHA.plot_queue[0], pgplot.PlotObject1D):
            fig, ax = pgplot._mkplot1d(PHA, ax=ax, **kwargs)
        elif isinstance(PHA.plot_queue[0], pgplot.PlotObject2D):
            fig, ax = pgplot._mkplot2d(PHA, ax=ax, **kwargs)
        else:
            raise TypeError("Plot holding area has invalid entries.")
    # Reset the PHA to being empty
    if empty_pha:
        pgplot._reset_pha()
    return (fig, ax)


def mkinset(pos: Tuple[float, float],
            size: Tuple[float, float],
            ax: Any = None,
            parent_corners: Tuple[str, str] = ("sw", "se"),
            inset_corners: Tuple[str, str] = ("sw", "se"),
            transform: Any = None,
            plot_options: Optional[dict] = None,
            inset_options: Optional[dict] = None,
            show_zoom: bool = True,
            ) -> Any:
    """
    Constructs an inset plot, abstracting away the matplotlib interface,
    which is especially clunky with inverted axes.
    """
    # Find the axes to draw on
    ax = ax or plt.gca()
    # Generate inset axes
    inset_ax = ax.inset_axes([*pos, *size], transform=transform)
    plot_options = plot_options or {}
    mkplot(ax=inset_ax, xlabel="", ylabel="", **plot_options)

    # Convert the strings to numbers
    def convert_corner(ax: Any, cornerstr: str, is_inset: bool):
        """
        For non-inverted axes, ne = 1, nw = 2, sw = 3, se = 4
        If x-axis is inverted, ne = 2, nw = 1, sw = 4, se = 3
        If y-axis is inverted, ne = 4, nw = 3, sw = 2, se = 1
        If both are inverted,  ne = 3, nw = 4, sw = 1, se = 2
        However, inset axes ALWAYS behave as if they are not inverted,
        regardless of whether they were inverted or not.
        If there's a more elegant way of writing this, please let me know.
        """
        xinv, yinv = ax.xaxis_inverted(), ax.yaxis_inverted()
        if cornerstr in ["ne", "northeast"]:
            return 1 if (is_inset or (not xinv and not yinv)) else \
                2 if (xinv and not yinv) else \
                4 if (not xinv and yinv) else 3
        elif cornerstr in ["nw", "northwest"]:
            return 2 if (is_inset or (not xinv and not yinv)) else \
                1 if (xinv and not yinv) else \
                3 if (not xinv and yinv) else 4
        elif cornerstr in ["sw", "southwest"]:
            return 3 if (is_inset or (not xinv and not yinv)) else \
                4 if (xinv and not yinv) else \
                2 if (not xinv and yinv) else 1
        elif cornerstr in ["se", "southeast"]:
            return 4 if (is_inset or (not xinv and not yinv)) else \
                3 if (xinv and not yinv) else \
                1 if (not xinv and yinv) else 2
        else:
            raise ValueError("Invalid corner provided to mkinset().")

    from mpl_toolkits.axes_grid1.inset_locator import mark_inset  # type: ignore

    # The loc1 and loc2 are throwaway values, they'll be replaced later.
    default_inset_options = {"loc1": 1, "loc2": 2,
                             "edgecolor": "silver",
                             }
    # Construct options
    options = dict(default_inset_options)
    if inset_options is not None:
        options.update(inset_options)
    # Make the inset, if requested
    if show_zoom:
        _, line1, line2 = mark_inset(ax, inset_ax, **options)
        # Change the corners from which the lines are drawn.
        line1.loc1 = convert_corner(inset_ax, inset_corners[0], True)
        line1.loc2 = convert_corner(ax, parent_corners[0], False)
        line2.loc1 = convert_corner(inset_ax, inset_corners[1], True)
        line2.loc2 = convert_corner(ax, parent_corners[1], False)
    return inset_ax


def show(*args, **kwargs) -> None:
    return plt.show(*args, **kwargs)


def savefig(*args, **kwargs) -> None:
    return plt.savefig(*args, **kwargs)


def pause(*args, **kwargs) -> None:
    return plt.pause(*args, **kwargs)


def subplots(nrows=1, ncols=1, *args, **kwargs) -> Tuple[Any, Any]:
    # This implementation captures nrows and ncols so that we can set figsize
    # automatically. We don't care about the rest of the arguments, so those
    # can just be passed on directly.
    if "figsize" not in kwargs:
        kwargs["figsize"] = (ncols * 4, nrows * 4)
    return plt.subplots(nrows=nrows, ncols=ncols, *args, **kwargs)



from pathlib import Path

import penguins as pg

datadir = Path(__file__).parent.resolve() / "data"


def test_read_1D():
    # test 1D datasets
    path = datadir / "pt2"
    proton = pg.read(path, 1, 1)
    assert isinstance(proton, pg.dataset.Dataset1D)
    # make sure that strings are ok too
    proton = pg.read(str(path), 1, 1)
    assert isinstance(proton, pg.dataset.Dataset1D)


def test_read_1Dprojections():
    # test 1D projections
    path = datadir / "rot1"
    proj = pg.read(path, 8, 179)
    assert isinstance(proj, pg.dataset.Dataset1DProj)
    # make sure that strings are ok too
    proj = pg.read(str(path), 8, 179)
    assert isinstance(proj, pg.dataset.Dataset1DProj)


def test_read_2D():
    # test 2D datasets
    path = datadir / "rot1"
    hsqmbc = pg.read(path, 8, 1)
    assert isinstance(hsqmbc, pg.dataset.Dataset2D)
    # make sure that strings are ok too
    hsqmbc = pg.read(str(path), 8, 1)
    assert isinstance(hsqmbc, pg.dataset.Dataset2D)


def test_read_abs_1D():
    # test 1D datasets
    path = datadir / "pt2" / "1" / "pdata" / "1"
    proton = pg.read_abs(path)
    assert isinstance(proton, pg.dataset.Dataset1D)
    # make sure that strings are ok too
    proton = pg.read_abs(str(path))
    assert isinstance(proton, pg.dataset.Dataset1D)


def test_read_abs_1Dprojections():
    # test 1D projections
    path = datadir / "rot1" / "8" / "pdata" / "179"
    proj = pg.read_abs(path)
    assert isinstance(proj, pg.dataset.Dataset1DProj)
    # make sure that strings are ok too
    proj = pg.read_abs(str(path))
    assert isinstance(proj, pg.dataset.Dataset1DProj)


def test_read_abs_2D():
    # test 2D datasets
    path = datadir / "rot1" / "8" / "pdata" / "1"
    hsqmbc = pg.read_abs(path)
    assert isinstance(hsqmbc, pg.dataset.Dataset2D)
    # make sure that strings are ok too
    hsqmbc = pg.read_abs(str(path))
    assert isinstance(hsqmbc, pg.dataset.Dataset2D)


# Tests for the plotting functions are elsewhere.



from __future__ import annotations   # PEP 563

from itertools import zip_longest, cycle
from collections import abc
from typing import (Union, Iterable, Dict, Optional, Any,
                    Tuple, List, Deque, Callable, Sequence, Iterator)
from numbers import Real

import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from matplotlib.ticker import AutoMinorLocator  # type: ignore
from matplotlib.legend_handler import HandlerBase  # type: ignore
import seaborn as sns  # type: ignore

from . import dataset as ds
from . import main
from .type_aliases import *


# -- HELPER OBJECTS -------------------------------------------

# 1D colour palette to use (from seaborn). By default "deep".
_current_palette = "deep"
# This are the colours from seaborn-bright, but rearranged into nice tuples.
# Honestly three pairs of colours should suffice. If you're plotting more 2D
# spectra than that on the same graph, you probably need to rethink your plot,
# or at least choose your colours manually to illustrate whatever point you're
# trying to make.
_bright_2d = [("#023EFF", "#E8000B"), # blue, red
              ("#1AC938", "#FF7C00"), # green, orange
              ("#8B2BE2", "#F14CC1"), # purple, pink
              ]


def set_palette(palette: str
                ) -> None:
    global _current_palette
    # Change seaborn palette, in case user wants to draw other plots.
    sns.set_palette(palette)
    # Change penguins 1D palette.
    _current_palette = palette


class PlotHoldingArea():
    def __init__(self) -> None:
        self.plot_queue: List = []
        # prime the colour generators
        self.colors_1d = self.color_generator_1d()
        self.colors_2d = self.color_generator_2d()

    def color_generator_1d(self) -> Iterator[str]:
        yield from cycle(sns.color_palette(_current_palette))

    def color_generator_2d(self) -> Iterator[Tuple[str, str]]:
        yield from cycle(_bright_2d)

class PlotProperties():
    """
    A collection of properties of spectra that have already been constructed
    with mkplot().
    """
    def __init__(self) -> None:
        self.hoffsets: List[float] = []
        self.voffsets: List[float] = []
        self.colors: List[str] = []
        self.options: List[Dict[str, Any]] = []
        self.colors_positive: List[str] = []
        self.colors_negative: List[str] = []

# Not a good idea to use these directly!
_globalPHA: PlotHoldingArea = PlotHoldingArea()
_globalPP: PlotProperties = PlotProperties()

# Go through these methods! It doesn't necessarily make it *safer*, because
# you can still access _globalPHA, but it's easier to keep track of.
def get_pha() -> PlotHoldingArea:
    return _globalPHA

def _reset_pha() -> None:
    global _globalPHA
    _globalPHA = PlotHoldingArea()

def get_properties() -> PlotProperties:
    return _globalPP

def _reset_properties() -> None:
    global _globalPP
    _globalPP = PlotProperties()

# -- 1D PLOTTING ----------------------------------------------

class PlotObject1D():
    """
    Object that includes a Dataset as well as plotting options, e.g.
    labels, bounds, colours, & generic options passed to matplotlib.
    """
    default_1d_plotoptions = {"linewidth": 1}
    def __init__(self,
                 dataset: ds.TDataset1D,
                 scale: float = 1,
                 bounds: TBounds = "",
                 dfilter: Optional[Callable[[float], bool]] = None,
                 label: OS = None,
                 color: OS = None,
                 plot_options: Optional[Dict] = None
                 ) -> None:
        self.dataset = dataset
        self.scale = scale
        self.bounds = bounds
        self._init_options(plot_options, color, label)
        self.ppm_scale = self.dataset.ppm_scale(bounds=self.bounds)
        # Handle processed data
        proc_data = self.dataset.proc_data(bounds=self.bounds)
        if dfilter is not None:
            # This is a bit ugly, but if dfilter involves any ands or ors,
            # then proc_data = proc_data[dfilter(proc_data)] will raise
            # a ValueError: Truth value is ambiguous.
            logical_array = np.array([dfilter(i) for i in proc_data])
            proc_data = np.where(logical_array, proc_data, np.nan)
        self.proc_data = proc_data * self.scale

    def _init_options(self,
                      plot_options: Optional[Dict],
                      color: OS,
                      label: OS):
        """
        Note that the color parameter will override any color key/value pair
        passed in plot_options.
        """
        # Make a copy of the default options.
        options: Dict[str, Any] = dict(self.default_1d_plotoptions)
        # Update it using the function arguments
        if plot_options is not None:
            options.update(plot_options)
        if color is not None:
            options.update(color=color)
        if label is not None:
            options.update(label=label)
        # If a colour hasn't been chosen by now, get one from the colour
        # generator in PHA.
        if "color" not in options:
            next_color = next(get_pha().colors_1d)
            options.update(color=next_color)
        # Finally, set the instance attribute
        self.options = options


def stage1d(dataset: ds.TDataset1D,
            scale: float = 1,
            bounds: TBounds = "",
            dfilter: Optional[Callable[[float], bool]] = None,
            label: OS = None,
            color: OS = None,
            plot_options: Optional[Dict] = None,
            ) -> None:
    """
    Stages a 1D spectrum to be plotted, i.e. sticks it into PHA.plot_queue.
    """
    # Create a PlotObject1D first.
    plot_obj = PlotObject1D(dataset=dataset, scale=scale,
                            bounds=bounds, dfilter=dfilter,
                            label=label,
                            color=color, plot_options=plot_options)

    # Check that the plot queue doesn't have 2D spectra.
    # We can just check against the first element. By induction, it is
    # equivalent to checking against every element.
    PHA = get_pha()
    if len(PHA.plot_queue) != 0 and isinstance(PHA.plot_queue[0], PlotObject2D):
        raise TypeError("Plot queue already contains 2D spectra.")
    else:
        PHA.plot_queue.append(plot_obj)


def _mkplot1d(holding_area: PlotHoldingArea,
              ax: Any = None,
              style: str = "1d",
              stacked: bool = False,
              voffset: Union[Sequence, float] = 0,
              hoffset: Union[Sequence, float] = 0,
              title: OS = None,
              autolabel: OS = None,
              xlabel: OS = None,
              ylabel: OS = None,
              legend_loc: Any = "best",
              ) -> Tuple[Any, Any]:
    """
    Plots all the 1D objects in the PHA.
    """
    make_legend = False
    # Find the maximum height
    heights = [np.nanmax(pobj.proc_data) - np.nanmin(pobj.proc_data)
               for pobj in holding_area.plot_queue]
    max_height = max(heights)

    # Get Axes object
    if ax is None:
        ax = plt.gca()

    # Iterate over plot objects
    for n, pobj in enumerate(holding_area.plot_queue):
        # Calculate the hoffset and voffset for this spectrum. If offset is a
        # sequence then use offset[n], otherwise if it's a float use n * offset
        # Also, mypy really doesn't like try/except.
        try:
            this_hoffset = hoffset[n]   # type: ignore
        except TypeError:  # not a sequence
            this_hoffset = hoffset * n   # type: ignore
        try:
            this_voffset = voffset[n] * max_height   # type: ignore
        except TypeError:  # not a sequence
            if stacked:
                # Raise each spectrum by the heights of previous spectra, plus
                # a little padding per spectrum.
                this_voffset = sum(heights[0:n]) + (n * 0.1 * max_height)
            else:
                # This covers the case where voffset is 0 as well.
                this_voffset = n * voffset * max_height  # type: ignore
        # Decide whether to make the legend
        if "label" in pobj.options and pobj.options["label"] is not None:
            make_legend = True
        # Plot it!
        ax.plot(pobj.ppm_scale - this_hoffset,
                pobj.proc_data + this_voffset,
                **pobj.options)
        # Add heights and colors to plotproperties.
        pp = get_properties()
        pp.hoffsets.append(this_hoffset)
        pp.voffsets.append(this_voffset)
        pp.colors.append(pobj.options["color"])
        pp.options.append(pobj.options)

    # Figure out the x- and y-labels.
    # First, we generate the strings accoring to autolabel.
    f_ylabel = "Intensity (au)"
    if autolabel is not None:
        if autolabel == "nucl":
            f_xlabel = holding_area.plot_queue[0].dataset.nuclei_to_str()
            f_xlabel += " (ppm)"
        else:
            raise ValueError(f"Invalid value '{autolabel}' given for "
                             "parameter autolabel.")
    else:
        f_xlabel = "Chemical shift (ppm)"
    # Then we override them based on the values of mkplot()'s kwargs.
    f_xlabel = xlabel if xlabel is not None else f_xlabel
    f_ylabel = ylabel if ylabel is not None else f_ylabel

    # Plot formatting
    # Only if y-axis is enabled.
    # plt.ylabel(f_ylabel)
    # ax.ticklabel_format(axis='y', style='sci', scilimits=(-1, 4), useMathText=True)
    if title:
        ax.set_title(title)
    ax.set_xlabel(f_xlabel)
    if not ax.xaxis_inverted():
        ax.invert_xaxis()
    if make_legend:
        ax.legend(loc=legend_loc)
    # Apply axis styles.
    style_axes(ax, style)
    return plt.gcf(), ax


def style_axes(ax: Any,
               style: str,
               ) -> None:
    """
    Styles the Axes instance according to the given style. Useful for ensuring
    that all subplots have a homogenous look.
    """
    if style == "1d":
        # Default 1D style. Doesn't draw a bounding box, only has the bottom
        # spine, which is made thicker. Adds extra x-axis ticks.
        for s in ["top", "left", "right"]:
            ax.spines[s].set_visible(False)
        ax.yaxis.set_visible(False)
        # Make the bottom one thicker
        ax.spines["bottom"].set_linewidth(1.3)
        # Enable minor ticks and make the ticks more visible
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(which="both", width=1.3)
        ax.tick_params(which="major", length=5)
        ax.tick_params(which="minor", length=3)
        plt.tight_layout()
    elif style == "1d_box":
        # 1D but with thick bounding box. Extra x-axis ticks added. The y-axis
        # is still disabled.
        ax.yaxis.set_visible(False)
        # Make spines thicker
        for s in ["top", "left", "right", "bottom"]:
            ax.spines[s].set_linewidth(1.3)
        # Enable minor ticks and make the ticks more visible
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(which="both", width=1.3)
        ax.tick_params(which="major", length=5)
        ax.tick_params(which="minor", length=3)
        plt.tight_layout()
    elif style == "2d":
        # Default 2D style. Makes bounding box thicker and adds x- and y-axis
        # ticks.
        for s in ["top", "left", "right", "bottom"]:
            ax.spines[s].set_linewidth(1.3)
        # Enable minor ticks and make the ticks more visible
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(which="both", width=1.3)
        ax.tick_params(which="major", length=5)
        ax.tick_params(which="minor", length=3)
        plt.tight_layout()
    elif style == "plot":
        # To be used for other plots, e.g. seaborn / matplotlib visualisations.
        # Draws a thicker bounding box around the Axes, but otherwise doesn't
        # try to change axis ticks etc.
        for s in ["top", "left", "right", "bottom"]:
            ax.spines[s].set_linewidth(1.3)
        plt.tight_layout()
    elif style == "natural":
        # Literally, do nothing. Not even tight_layout().
        pass
    else:
        raise ValueError(f"Invalid style '{style}' requested.")


# -- 2D PLOTTING ----------------------------------------------

class Contours:
    """2D plot contours."""
    def __init__(self,
                 dataset: ds.Dataset2D,
                 levels: TLevels = (None, None, None),
                 colors: TColors = (None, None),
                 ) -> None:
        self.dataset = dataset
        if isinstance(levels, float):
            levels = (levels, None, None)
        self.make_levels(*levels)
        self.make_colors(*colors)

    def make_levels(self, base: OF = None,
                    increment: OF = None,
                    number: Optional[int] = None
                    ) -> None:
        self.base = base or self.dataset._tsbaselev
        self.increment = increment or 1.5
        self.number = number or 10

    def make_colors(self,
                    color_positive: OS = None,
                    color_negative: OS = None
                    ) -> None:
        if color_positive is None or color_negative is None:
            # Means we need to get a color set from the PHA generator.
            next_positive, next_negative = next(get_pha().colors_2d)
            self.color_positive = color_positive or next_positive
            self.color_negative = color_negative or next_negative
        else:
            self.color_positive = color_positive
            self.color_negative = color_negative

    def generate_contour_levels(self) -> List[float]:
        neg = [-self.base * (self.increment ** (self.number - i)) for i in range(self.number)]
        pos = [self.base * (self.increment ** i) for i in range(self.number)]
        return neg + pos

    def generate_contour_colors(self) -> List[str]:
        neg = [self.color_negative] * self.number
        pos = [self.color_positive] * self.number
        return neg + pos


class PlotObject2D():
    """
    Object that includes a Dataset as well as plotting options, e.g.
    contours, bounds, colours, & generic options passed to matplotlib.
    """
    default_2d_plotoptions = {"linewidths": 0.7}
    def __init__(self,
                 dataset: ds.Dataset2D,
                 f1_bounds: TBounds = "",
                 f2_bounds: TBounds = "",
                 levels: TLevels = (None, None, None),
                 colors: TColors = (None, None),
                 dfilter: Optional[Callable[[float], bool]] = None,
                 label: OS = None,
                 plot_options: Optional[Dict] = None
                 ) -> None:
        self.dataset = dataset
        self.f1_bounds = f1_bounds
        self.f2_bounds = f2_bounds
        self.contours = Contours(self.dataset, levels, colors)
        # can access cpos and cneg with self.contours.color_[positive|negative]
        self.clevels = self.contours.generate_contour_levels()
        self.ccolors = self.contours.generate_contour_colors()
        self._init_options(plot_options)
        self.label = label
        # self.options will include the colors key.
        self.f1_scale = self.dataset.ppm_scale(axis=0, bounds=self.f1_bounds)
        self.f2_scale = self.dataset.ppm_scale(axis=1, bounds=self.f2_bounds)
        # Handle processed data
        proc_data = self.dataset.proc_data(f1_bounds=self.f1_bounds,
                                           f2_bounds=self.f2_bounds)
        if dfilter is not None:
            logical_array = np.array([dfilter(i) for i in proc_data.flat])
            logical_array = np.reshape(logical_array, np.shape(proc_data))
            proc_data = np.where(logical_array, proc_data, np.nan)
        self.proc_data = proc_data

    def _init_options(self,
                      plot_options: Optional[Dict],
                      ) -> None:
        # Make a copy of the default options
        options: Dict[str, Any] = dict(self.default_2d_plotoptions)
        if plot_options is not None:
            options.update(plot_options)
        # As before, the colors parameter will override the colors key provided
        # in plot_options.
        options.update(colors=self.ccolors)
        self.options = options


class ContourLegendHandler(HandlerBase):
    # code lifted mostly from https://stackoverflow.com/a/41765095/
    def create_artists(self, legend, orig_handle,
                       x0, y0, width, height, fontsize, trans):
        line_cpos = plt.Line2D([x0, y0 + width], [0.7 * height, 0.7 * height],
                               color=orig_handle[0])
        line_cneg = plt.Line2D([x0, y0 + width], [0.3 * height, 0.3 * height],
                               color=orig_handle[1])
        return [line_cpos, line_cneg]


def stage2d(dataset: ds.Dataset2D,
            f1_bounds: TBounds = "",
            f2_bounds: TBounds = "",
            levels: TLevels = (None, None, None),
            colors: TColors = (None, None),
            dfilter: Optional[Callable[[float], bool]] = None,
            label: OS = None,
            plot_options: Optional[Dict] = None,
            ) -> None:
    """
    Stages a 2D spectrum.
    """
    plot_obj = PlotObject2D(dataset=dataset,
                            f1_bounds=f1_bounds, f2_bounds=f2_bounds,
                            levels=levels, colors=colors,
                            dfilter=dfilter, label=label,
                            plot_options=plot_options)

    # Check that the plot queue doesn't have 1D spectra.
    # We can just check against the first element. By induction, it is
    # equivalent to checking against every element.
    PHA = get_pha()
    if len(PHA.plot_queue) != 0 and isinstance(PHA.plot_queue[0], PlotObject1D):
        raise TypeError("Plot queue already contains 1D spectra.")
    else:
        PHA.plot_queue.append(plot_obj)


def _mkplot2d(holding_area: PlotHoldingArea,
              ax: Any = None,
              style: str = "2d",
              offset: Tuple[float, float] = (0, 0),
              title: OS = None,
              autolabel: OS = None,
              xlabel: OS = None,
              ylabel: OS = None,
              legend_loc: Any = "best",
              ) -> Tuple[Any, Any]:
    """
    Plots all the 2D objects in the PHA.
    """
    make_legend = False
    legend_colors, legend_labels = [], []
    if ax is None:
        ax = plt.gca()
    # Iterate over plot objects
    for n, pobj in enumerate(holding_area.plot_queue):
        ax.contour(pobj.f2_scale - (n * offset[1]),   # x-axis
                   pobj.f1_scale - (n * offset[0]),   # y-axis
                   pobj.proc_data,
                   levels=pobj.clevels,
                   **pobj.options)
        # Construct lists for plt.legend
        if pobj.label is not None:
            make_legend = True
            legend_colors.append((pobj.contours.color_positive,
                                  pobj.contours.color_negative)
                                 )
            legend_labels.append(pobj.label)

    # Figure out the x- and y-labels.
    # First, we generate the strings accoring to autolabel.
    if autolabel is not None:
        if autolabel == "nucl":
            f_xlabel = holding_area.plot_queue[0].dataset.nuclei_to_str()[1]
            f_xlabel += " (ppm)"
            f_ylabel = holding_area.plot_queue[0].dataset.nuclei_to_str()[0]
            f_ylabel += " (ppm)"
        else:
            raise ValueError(f"Invalid value '{autolabel}' given for "
                             "parameter autolabel.")
    else:
        f_xlabel = r"$f_2$ (ppm)"
        f_ylabel = r"$f_1$ (ppm)"
    # Then we override them based on the values of mkplot()'s kwargs.
    f_xlabel = xlabel if xlabel is not None else f_xlabel
    f_ylabel = ylabel if ylabel is not None else f_ylabel

    # Plot formatting
    # Only if y-axis is enabled.
    # plt.ylabel(f_ylabel)
    # ax.ticklabel_format(axis='y', style='sci', scilimits=(-1, 4), useMathText=True)
    # Axis formatting
    if not ax.xaxis_inverted():
        ax.invert_xaxis()
    if not ax.yaxis_inverted():
        ax.invert_yaxis()
    if title is not None:
        ax.set_title(title)
    ax.set_xlabel(f_xlabel)
    ax.set_ylabel(f_ylabel)
    # Apply other styles.
    style_axes(ax, style)

    # Make legend. This part is not easy...
    # See https://stackoverflow.com/questions/41752309/ for an example
    if make_legend:
        plt.legend(legend_colors, legend_labels,
                   handler_map={tuple: ContourLegendHandler()},
                   loc=legend_loc)
    return (plt.gcf(), ax)


def _make_contour_slider(dataset: ds.Dataset2D,
                         increment: float = None,
                         nlev: int = 4,
                         ) -> float:
    # Choose contour levels. We reduce nlev to 4 by default so that the
    # plotting is faster -- otherwise it's super laggy. We try to cover the
    # same dynamic range as 1.5 ** 10 by default, unless the user specified
    # an increment.
    initial_baselev = dataset._tsbaselev
    increment = increment or (1.5 ** 10) ** (1 / nlev)
    initial_clev = (initial_baselev, increment, nlev)
    dataset.stage(levels=initial_clev)
    # Maximum level should be the highest intensity of the spectrum.
    max_baselev = np.max(np.abs(dataset.rr))

    # Plot the spectrum on the top portion of the figure.
    fig, ax = plt.subplots()
    _, plot_axes = main.mkplot(empty_pha=False,
                               style="natural",
                               )
    orig_xlim = plot_axes.get_xlim()
    orig_ylim = plot_axes.get_ylim()
    plt.subplots_adjust(left=0.1, bottom=0.25)

    # Generate a slider.
    from matplotlib.widgets import Slider, Button     # type: ignore
    baselev_axes = plt.axes([0.1, 0.1, 0.8, 0.03], facecolor="lavender")
    baselev_slider = Slider(baselev_axes, "",
                            2, np.log10(max_baselev),
                            valinit=np.log10(initial_baselev), color="purple")
    # Add some text
    plt.text(0.5, -1.2, r"log$_{10}$(base contour level)",
             horizontalalignment="center",
             transform=baselev_axes.transAxes)

    # Define the behaviour when redrawn
    def redraw(plot_axes: Any,
               val: float
               ) -> None:
        # Update the internal Contours object
        pobj = get_pha().plot_queue[0]
        pobj.contours.base = 10 ** val
        # Regenerate the contours
        pobj.clevels = pobj.contours.generate_contour_levels()
        # Replot
        xlim = plot_axes.get_xlim()
        ylim = plot_axes.get_ylim()
        plot_axes.cla()
        plt.sca(plot_axes)
        main.mkplot(close=False, empty_pha=False, style="natural")
        plot_axes.set_xlim(xlim)
        plot_axes.set_ylim(ylim)

    # Register the redraw function. The argument to on_changed must be a
    # function taking val as its only parameter.
    baselev_slider.on_changed(lambda val: redraw(plot_axes, val))

    # Generate the "OK" button
    okay_axes = plt.axes([0.84, 0.025, 0.1, 0.04])
    okay_button = Button(okay_axes, "OK", color="plum", hovercolor='0.95')
    okay = 0
    # Add the close behaviour
    def set_okay(button):
        nonlocal okay
        okay = 1
        plt.close()
    okay_button.on_clicked(set_okay)

    # Generate the "Reset zoom" button
    reset_axes = plt.axes([0.66, 0.025, 0.16, 0.04])
    reset_button = Button(reset_axes, "Reset zoom", color="plum", hovercolor='0.95')
    # Add the reset behaviour
    def reset(button):
        plot_axes.set_xlim(orig_xlim)
        plot_axes.set_ylim(orig_ylim)
    reset_button.on_clicked(reset)

    # Show it to the user.
    main.show()

    # Post-OK actions
    if okay:
        # Print the dataset
        print(f"\n[[{dataset!r}]]")
        # Print the initial value
        ival = initial_baselev
        ival_short = f"{ival:.2e}".replace("e+0", "e")
        print(f"The initial base level was: {ival:.2f} (or {ival_short})")
        # Print the final value
        fval = 10 ** baselev_slider.val
        fval_short = f"{fval:.2e}".replace("e+0", "e")
        print(f"The final base level was:   {fval:.2f} (or {fval_short})\n")
    # Clear out the PHA
    _reset_pha()
    plt.close("all")

    return fval if okay else None



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="penguins",
    version="0.1",
    author="Jonathan Yong",
    author_email="yongrenjie@gmail.com",
    description="Penguins: an Easy, NPE-free Gateway to Unpacking and Illustrating NMR Spectra",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yongrenjie/penguins",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "numpy>=1.17.0",
        "pandas",
        "matplotlib",
        "seaborn",
    ]

)



from __future__ import annotations

from collections import UserDict, abc
from pathlib import Path
from typing import (Any, Union, Tuple, Optional,
                    TypeVar, Callable, overload)

import numpy as np               # type: ignore
import matplotlib.pyplot as plt  # type: ignore

from . import pgplot
from .type_aliases import *


def _try_convert(x: Any, type: Any):
    # Tries to convert an input to a specific type.
    # Returns the original input if it fails.
    # Works with tuples and lists too: tries to convert individual elements.
    # Works with np.ndarrays using ndarray.astype()
    try:
        return type(x)
    except ValueError:
        return x
    except TypeError:
        if isinstance(x, tuple) or isinstance(x, list):
            y = [_try_convert(i, type) for i in x]
            if isinstance(x, tuple):
                return tuple(y)
            else:
                return y
        elif isinstance(x, np.ndarray):
            return x.astype(dtype=type)
        else:
            return x


class _parDict(UserDict):
    """Dictionary for storing acquisition & processing parameters.

    Initialise within a Dataset class as self.pars = _parDict(self.path).

    Parameter names are stored as lower-case strings. When looking up a
    parameter, if its value is not already stored, the dictionary will look
    it up in the associated TopSpin parameter files. Subsequently, the value
    is cached.

    Therefore, the dictionary can be treated *as if* it were already fully
    populated when initialised; but this avoids the time spent parsing the
    entire file for parameters, the vast majority of which are useless."""

    def _editkey(self, key: object):
        """Method for converting keys before lookup."""
        return str(key).lower()

    def __init__(self, path: Path):
        self.path = path
        self._p_acqus = path.parents[1] / "acqus"
        self._p_procs = path / "procs"
        self._p_acqu2s = path.parents[1] / "acqu2s" \
                         if (path.parents[1] / "acqu2s").exists() else None
        self._p_proc2s = path / "proc2s" \
                         if (path / "proc2s").exists() else None
        super().__init__()

    def _getpar(self, par: str):
        # Get the direct dimension
        val = self._get_acqus_par(par, self._p_acqus)
        if val is None:
            val = self._get_procs_par(par, self._p_procs)
        val = _try_convert(val, float)
        # Check if there's also an indirect dimension
        if self._p_acqu2s is not None:
            val_indirect = self._get_acqus_par(par, self._p_acqu2s)
            if val_indirect is None and self._p_proc2s is not None:
                val_indirect = self._get_procs_par(par, self._p_proc2s)
            val_indirect = _try_convert(val_indirect, float)
            # Make a tuple if the indirect dimension was found
            # We make an exception for NS because for some reason, TopSpin stores
            # an "NS" in the indirect dimension which is just 1.
            if val_indirect is not None and par.lower() != "ns":
                if isinstance(val, float) and isinstance(val_indirect, float):
                    # ndarrays are good, so that we can do elementwise stuff
                    # like self["o1p] = self["o1"] / self["bf1"]
                    val = np.array([val_indirect, val])
                else:
                    val = (val_indirect, val)
        # Some parameters must be ints, this will save the user (and me) headaches
        int_pars = ["td", "si", "ns", "xdim", "nustd", "bytorda", "bytordp",
                    "dtypa", "dtypp"]
        if par.lower() in int_pars:
            val = _try_convert(val, int)
        return val

    def _get_acqus_par(self, par: str, fp: Union[str, Path]):
        """Get the value of an acquisition parameter and return as string."""
        # Capitalise and remove any spaces from par
        par = par.upper()
        if len(par.split()) > 1:
            par = "".join(par.split())
        # Split par into number-less bit and number bit
        parl = par.rstrip("1234567890")
        parr = par[len(parl):]
        params_with_space = ["CNST", "D", "P", "PLW", "PCPD", "GPX", "GPY",
                             "GPZ", "SPW", "SPOAL", "SPOFFS", "L", "IN",
                             "INP", "PHCOR"]
        # Get the parameter
        if (parr != "") and (parl in params_with_space):  # e.g. cnst2
            with open(fp, "r") as file:
                # Read up to the line declaring the parameters
                for line in file:
                    if line.upper().startswith(f"##${parl}="):
                        break
                else:   # triggers if didn't break -- i.e. parameter was not found
                    return None
                # Grab the values and put them in a list
                s = ""
                # Read until next parameter
                line = file.readline()
                while not line.startswith("##"):
                    s = s + line + " "
                    line = file.readline()
                # Pick out the desired value and return it
                return s.split()[int(parr)]
        else:                                             # e.g. sfo1 or rga
            with open(fp, "r") as file:
                for line in file:
                    if line.upper().startswith(f"##${par}="):
                        val = line.split(maxsplit=1)[-1].strip()
                        # strip away surrounding angle brackets
                        if val[0] == '<' and val[-1] == '>':
                            val = val[1:-1]
                        return val
        # If it hasn't been found
        return None

    def _get_procs_par(self, par: str, fp: Union[str, Path]):
        """Get the value of an processing parameter and return as string."""
        # Capitalise and remove any spaces from par
        par = par.upper()
        if len(par.split()) > 1:
            par = "".join(par.split())
        # Get the value (for processing parameters there aren't any lists like
        # CNST/D/P)
        with open(fp, "r") as file:
            for line in file:
                if line.upper().startswith(f"##${par}="):
                    val = line.split(maxsplit=1)[-1].strip()
                    # strip away surrounding angle brackets
                    if val[0] == '<' and val[-1] == '>':
                        val = val[1:-1]
                    return val

    def __getitem__(self, key: object):
        k = self._editkey(key)
        if k in self.data:
            return self.data[k]
        else:
            val = self._getpar(k)
            if val is not None:
                self.data[k] = val
                return val
            else:
                raise KeyError(key)

    def __setitem__(self, key: object, val: Any):
        k = self._editkey(key)
        self.data.__setitem__(k, val)

    def __delitem__(self, key: object):
        k = self._editkey(key)
        self.data.__delitem__(k)

    def __contains__(self, key: object):
        k = self._editkey(key)
        return k in self.data

    def __repr__(self):
        keys = sorted(list(self))
        s = "{" + ",\n".join(f"'{k}': {self[k]}" for k in keys) + "}"
        return s


def _parse_bounds(b: TBounds = "",
                  ) -> Tuple[OF, OF]:
    if isinstance(b, str):
        if b == "":
            return None, None
        elif b.startswith(".."):   # "..5" -> (None, 5)
            return None, float(b[2:])
        elif b.endswith(".."):   # "3.." -> (3, None)
            return float(b[:-2]), None
        elif ".." in b:
            x, y = b.split("..")
            xf, yf = float(x), float(y)  # let TypeError propagate
            if xf >= yf:
                raise ValueError(f"Use '{yf}..{xf}', not '{xf}..{yf}'.")
            return xf, yf
        else:
            raise ValueError(f"Invalid value {b} provided for bounds.")
    else:
        if len(b) != 2:
            raise ValueError(f"Invalid value {b} provided for bounds.")
        elif b[0] is not None and b[1] is not None and b[0] > b[1]:
            raise ValueError(f"Please use {(b[1], b[0])}, not {b}.")
        else:
            return b[0], b[1]

# -- Fundamental Dataset methods ------------------------

class _Dataset():
    """
    Defines behaviour that is common to all datasets.
    """

    def __init__(self,
                 path: Union[str, Path],
                 **kwargs
                 ) -> None:
        # Set up file path information
        self.path = Path(path).resolve().expanduser()
        self.expno = self.path.parents[1].name
        self.procno = self.path.name
        # Initialise _parDict and some key parameters
        self._initialise_pars()
        # Get paths to data and parameter files
        self._find_raw_data_paths()     # type: ignore # mixin
        self._find_proc_data_paths()    # type: ignore # mixin
        self._find_param_file_paths()   # type: ignore # mixin
        # Read in the spectral data
        self._read_raw_data()           # type: ignore # mixin
        self._read_spec()               # type: ignore # mixin

    def _initialise_pars(self) -> None:
        self.pars = _parDict(self.path)
        self["aq"] = (self["td"] / 2) / (self["sw"] * self["sfo1"])  # This is sfo1
        self["dw"] = self["aq"] * 1000000 / self["td"]
        self["o1p"] = self["o1"] / self["bf1"]   # This is BF1
        self["si"]
        self["nuc1"]
        # print(self.pars)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.path}')"

    def __getitem__(self, par: str):
        return self.pars[par]

    def __setitem__(self, par: str, val: Any):
        self.pars.__setitem__(par, val)

    def __delitem__(self, par:str):
        self.pars.__delitem__(par)


# -- 1D mixins ------------------------------------------

class _1D_RawDataMixin():
    def _find_raw_data_paths(self) -> None:
        self.path: Path
        self._p_fid = self.path.parents[1] / "fid"

    def _find_param_file_paths(self) -> None:
        self._p_procs = self.path / "procs"
        self._p_acqus = self.path.parents[1] / "acqus"

    def _read_raw_data(self) -> np.ndarray:
        datatype = "<" if self["bytorda"] == 0 else ">"  # type: ignore # mixin
        datatype += "i4"
        fid = np.fromfile(self._p_fid, dtype=datatype)
        fid = fid.reshape(int(self["td"]/2), 2)          # type: ignore # mixin
        fid = np.transpose(fid) * (2 ** self["nc"])      # type: ignore # mixin
        self.fid = fid[0] + (1j * fid[1])

    def raw_data(self) -> np.ndarray:
        return self.fid


class _1D_ProcDataMixin():

    def _find_proc_data_paths(self) -> None:
        self._p_real = self.path / "1r"        # type: ignore # mixin
        if (self.path / "1i").exists():        # type: ignore # mixin
            self._p_imag = self.path / "1i"    # type: ignore # mixin

    def _read_spec(self) -> np.ndarray:
        if self["dtypp"] == 0:                            # type: ignore # mixin
            dt = "<" if self["bytordp"] == 0 else ">"     # type: ignore # mixin
            dt += "i4"
        else:
            raise NotImplementedError("float data not yet accepted")
        self.real = np.fromfile(self._p_real, dtype=np.dtype(dt)) * (2 ** self["nc_proc"])  # type: ignore # mixin
        if hasattr(self, "_p_imag"):
            self.imag = np.fromfile(self._p_imag, dtype=np.int32) * (2 ** self["nc_proc"])  # type: ignore # mixin

    def proc_data(self,
                  bounds: TBounds = "",
                  ) -> np.ndarray:
        return self.real[self._ppm_to_slice(bounds)]

    def _ppm_to_slice(self,
                      bounds: TBounds = "",
                      ) -> slice:
        """
        Converts a tuple of chemical shifts (upper, lower) to a slice object.
        Note that upper must be greater than lower. Either upper or lower can
        be None, in which case the spectrum limit is used.
        """
        lower, upper = _parse_bounds(bounds)
        return slice(self._ppm_to_index(upper) or 0,                      # type: ignore # mixin
                     (self._ppm_to_index(lower) or self["si"] - 1) + 1)   # type: ignore # mixin

    def integrate(self,
                  peak: OF = None,
                  margin: OF = None,
                  mode: str = "sum",
                  bounds: TBounds = None,
                  ) -> float:
        """
        Integrates a region of a spectrum.
        Regions can either be defined via peak and margin, which leads to the region of
        (peak - margin) to (peak + margin), or manually via the bounds parameter. Note that
        specifying (peak, margin) overrules the bounds parameter if both are passed.

        Default mode is "sum", which simply adds up all points in the region. Alternatives are
        "max" which returns the highest peak, and "min" which returns the lowest peak.
        """
        integration_functions = {"sum": np.sum, "max": np.amax, "min": np.amin}
        if mode not in integration_functions:
            raise ValueError(f"Invalid value '{mode}' specified for integration mode.")
        # Process (peak, margin)
        if peak is not None and margin is not None:
            bounds = f"{peak-margin}..{peak+margin}"
        # We will run into this if user has not specified enough information
        if bounds is None:
            raise ValueError("Please specify integration region.")
        # Perform the integration
        finteg = integration_functions[mode]
        return finteg(self.proc_data(bounds))

    def nuclei_to_str(self
                      ) -> str:
        """
        Returns a string with the nucleus nicely formatted for use as the
        xlabel in a plot.
        """
        nuc = self["nuc1"]          # type: ignore
        elem = nuc.lstrip("1234567890")
        mass = nuc[:-len(elem)]
        return rf"$^{{{mass}}}${elem}"


class _1D_PlotMixin():

    def stage(self, *args, **kwargs) -> None:
        pgplot.stage1d(self, *args, **kwargs)  # type: ignore # mixin


# -- 2D mixins ------------------------------------------

class _2D_RawDataMixin():

    def _find_raw_data_paths(self) -> None:
        self.path: Path
        self._p_ser = self.path.parents[1] / "ser"

    def _find_param_file_paths(self) -> None:
        self._p_procs = self.path / "procs"
        self._p_acqus = self.path.parents[1] / "acqus"
        self._p_proc2s = self.path / "proc2s"
        self._p_acqu2s = self.path.parents[1] / "acqu2s"

    def _read_raw_data(self) -> np.ndarray:
        if self["dtypa"] == 0:                             # type: ignore # mixin
            dtype = "<" if self["bytorda"] == 0 else ">"   # type: ignore # mixin
            dtype += "i4"
        else:
            raise NotImplementedError("float data not yet accepted")
        # TODO
        # ser = np.fromfile(self._p_ser, dtype=??)
        pass

    def raw_data(self):
        # TODO
        raise NotImplementedError


class _2D_ProcDataMixin():

    def _find_proc_data_paths(self) -> None:
        self.path: Path
        self._p_rr = self.path / "2rr"
        if (self.path / "2ri").exists():
            self._p_ri = self.path / "2ri"
        if (self.path / "2ir").exists():
            self._p_ir = self.path / "2ir"
        if (self.path / "2ii").exists():
            self._p_ii = self.path / "2ii"

    def _read_spec(self) -> np.ndarray:
        # Helper function
        def _read_one_spec(spec_path = Path) -> np.ndarray:
            if np.all(self["dtypp"] == 0):                           # type: ignore # mixin
                dt = "<" if np.all(self["bytordp"] == 0) else ">"    # type: ignore # mixin
                dt += "i4"
            else:
                raise NotImplementedError("float data not yet accepted")
            sp = np.fromfile(spec_path, dtype=np.dtype(dt))
            sp = sp.reshape(self["si"])                              # type: ignore # mixin
            # Format according to xdim. See TopSpin "data format" manual.
            # See also http://docs.nmrfx.org/viewer/files/datasets.
            nrows, ncols = self["si"] / self["xdim"]                 # type: ignore # mixin
            submatrix_size = np.prod(self["xdim"])                   # type: ignore # mixin
            nrows = int(nrows); ncols = int(ncols)
            sp = sp.reshape(self["si"][0] * ncols, self["xdim"][1])  # type: ignore # mixin
            sp = np.vsplit(sp, nrows * ncols)
            sp = np.concatenate(sp, axis=1)
            sp = np.hsplit(sp, nrows)
            sp = np.concatenate(sp, axis=0)
            sp = sp.reshape(self["si"])                              # type: ignore # mixin
            return sp * (2 ** self["nc_proc"][1])                    # type: ignore # mixin

        # Read the spectra
        self.rr = _read_one_spec(self._p_rr)
        # Calculate a suitable baselev
        self._tsbaselev = self["s_dev"][1] * 35 * (2 ** self["nc_proc"][1])    # type: ignore # mixin
        if hasattr(self, "_p_ri"):
            self.ri = _read_one_spec(self._p_ri)
        if hasattr(self, "_p_ir"):
            self.ir = _read_one_spec(self._p_ir)
        if hasattr(self, "_p_ii"):
            self.ii = _read_one_spec(self._p_ii)

    def proc_data(self,
                  f1_bounds: TBounds = "",
                  f2_bounds: TBounds = "",
                  ) -> np.ndarray:
        f1_slice = self._ppm_to_slice(axis=0, bounds=f1_bounds)
        f2_slice = self._ppm_to_slice(axis=1, bounds=f2_bounds)
        return self.rr[f1_slice, f2_slice]

    def _ppm_to_slice(self,
                      axis: int,
                      bounds: TBounds = "",
                      ) -> slice:
        """
        Converts a tuple of chemical shifts (upper, lower) to a slice object.
        Note that upper must be greater than lower. Either upper or lower can
        be None, in which case the spectrum limit is used.
        Axis = 0 for f1, 1 for f2.
        """
        lower, upper = _parse_bounds(bounds)
        return slice(self._ppm_to_index(axis, upper) or 0,                           # type: ignore # mixin
                     (self._ppm_to_index(axis, lower) or self["si"][axis] - 1) + 1)  # type: ignore # mixin

    def integrate(self,
                  peak: Optional[Tuple[float, float]] = None,
                  margin: Optional[Tuple[float, float]] = None,
                  mode: str = "sum",
                  f1_bounds: TBounds = None,
                  f2_bounds: TBounds = None,
                  ) -> float:
        """
        Integrates a region of a spectrum.
        Regions can either be defined via peak and margin, which leads to the region of
        (peak - margin) to (peak + margin) in both dimensions, or manually via the (f1,f2)_bounds
        parameter. Note that specifying (peak, margin) overrules the bounds parameter if both are passed.

        Default mode is "sum", which simply adds up all points in the region. Alternatives are
        "max" which returns the highest peak, and "min" which returns the lowest peak.
        """
        integration_functions = {"sum": np.sum, "max": np.amax, "min": np.amin}
        if mode not in integration_functions:
            raise ValueError(f"Invalid value '{mode}' specified for integration mode.")
        # Process (peak, margin)
        if peak is not None and margin is not None:
            f1_bounds = f"{peak[0]-margin[0]}..{peak[0]+margin[0]}"
            f2_bounds = f"{peak[1]-margin[1]}..{peak[1]+margin[1]}"
        # We will run into this if user has not specified enough information
        if f1_bounds is None or f2_bounds is None:
            raise ValueError("Please specify integration region.")
        # Perform the integration
        finteg = integration_functions[mode]
        return finteg(self.proc_data(f1_bounds, f2_bounds))

    def nuclei_to_str(self,
                      ) -> Tuple[str, str]:
        """
        Returns a tuple of two strings with the nuclei nicely formatted for use
        as xlabels or ylabels in plots.
        """
        f1, f2 = self["nuc1"]              # type: ignore
        f1_elem = f1.lstrip("1234567890")
        f1_mass = f1[:-len(f1_elem)]
        f2_elem = f2.lstrip("1234567890")
        f2_mass = f2[:-len(f2_elem)]
        return (rf"$^{{{f1_mass}}}${f1_elem}", rf"$^{{{f2_mass}}}${f2_elem}")


class _2D_PlotMixin():

    def stage(self, *args, **kwargs) -> None:
        pgplot.stage2d(self, *args, **kwargs)  # type: ignore # mixin

    def find_baselev(self, *args, **kwargs):
        pgplot._make_contour_slider(self, *args, **kwargs)


# -- Actual dataset classes -----------------------------

class Dataset1D(_1D_RawDataMixin,
                _1D_ProcDataMixin,
                _1D_PlotMixin,
                _Dataset):

    def _ppm_to_index(self,
                      ppm: Optional[float]
                      ) -> Optional[int]:
        """
        Finds the index of the spectrum which is closest to the given
        chemical shift. Returns None if ppm is None.
        """
        if ppm is None:
            return None
        max_ppm = self["o1p"] + self["sw"]/2
        min_ppm = self["o1p"] - self["sw"]/2
        if ppm > max_ppm or ppm < min_ppm:
            raise ValueError(f"Chemical shift {ppm} is outside spectral window.")
        spacing = (max_ppm - min_ppm)/(self["si"] - 1)
        x = 1 + round((max_ppm - ppm)/spacing)
        return int(x)

    def ppm_scale(self,
                  bounds: TBounds = "",
                  ) -> np.ndarray:
        max_ppm = self["o1p"] + self["sw"]/2
        min_ppm = self["o1p"] - self["sw"]/2
        full_scale = np.linspace(max_ppm, min_ppm, self["si"])
        return full_scale[self._ppm_to_slice(bounds)]

    def hz_scale(self) -> np.ndarray:
        # These use SFO, not BF
        max_hz = self["o1"] + self["sw"]/(2 * self["sfo1"])
        min_hz = self["o1"] - self["sw"]/(2 * self["sfo1"])
        return np.linspace(max_hz, min_hz, self["si"])


class Dataset1DProj(_2D_RawDataMixin,
                    _1D_ProcDataMixin,
                    _1D_PlotMixin,
                    _Dataset):

    def _initialise_pars(self):
        # Initialise _parDict and common pars as before.
        super()._initialise_pars()
        # Then figure out which axis we have been projected onto.
        projdimen = self.pars._get_acqus_par("curexp",
                                             self.path / "used_from")
        if projdimen == "column":   # projection onto f1
            self.proj_axis = 0
        elif projdimen == "row":    # projection onto f2
            self.proj_axis = 1
        else:
            raise ValueError("Projection dimension was not found in "
                             "'used_from' file.")

    def _ppm_to_index(self,
                      ppm: Optional[float]
                      ) -> Optional[int]:
        """
        Finds the index of the spectrum which is closest to the given
        chemical shift. Returns None if ppm is None.
        """
        if ppm is None:
            return None
        max_ppm = self["o1p"][self.proj_axis] + self["sw"][self.proj_axis]/2
        min_ppm = self["o1p"][self.proj_axis] - self["sw"][self.proj_axis]/2
        if ppm > max_ppm or ppm < min_ppm:
            raise ValueError(f"Chemical shift {ppm} is outside spectral window.")
        spacing = (max_ppm - min_ppm)/(self["si"] - 1)
        x = 1 + round((max_ppm - ppm)/spacing)
        return int(x)

    def ppm_scale(self,
                  bounds: TBounds = "",
                  ) -> np.ndarray:
        max_ppm = self["o1p"][self.proj_axis] + self["sw"][self.proj_axis]/2
        min_ppm = self["o1p"][self.proj_axis] - self["sw"][self.proj_axis]/2
        full_scale = np.linspace(max_ppm, min_ppm, self["si"])
        return full_scale[self._ppm_to_slice(bounds)]

    def hz_scale(self) -> np.ndarray:
        # These use SFO, not BF
        max_hz = self["o1"][self.proj_axis] + self["sw"][self.proj_axis]/(2 * self["sfo1"][self.proj_axis])
        min_hz = self["o1"][self.proj_axis] - self["sw"][self.proj_axis]/(2 * self["sfo1"][self.proj_axis])
        return np.linspace(max_hz, min_hz, self["si"])


class Dataset2D(_2D_RawDataMixin,
                _2D_ProcDataMixin,
                _2D_PlotMixin,
                _Dataset):

    def _ppm_to_index(self,
                      axis: int,
                      ppm: Optional[float]
                      ) -> Optional[int]:
        """
        Finds the index of the spectrum which is closest to the given
        chemical shift. Returns None if ppm is None.
        Axis = 0 for f1, 1 for f2.
        """
        if ppm is None:
            return None
        max_ppm = self["o1p"][axis] + self["sw"][axis]/2
        min_ppm = self["o1p"][axis] - self["sw"][axis]/2
        if ppm > max_ppm or ppm < min_ppm:
            raise ValueError(f"Chemical shift {ppm} is outside spectral window.")
        spacing = (max_ppm - min_ppm)/(self["si"][axis] - 1)
        x = 1 + round((max_ppm - ppm)/spacing)
        return int(x)

    def ppm_scale(self,
                  axis: int,
                  bounds: TBounds = "",
                  ) -> np.ndarray:
        max_ppm = self["o1p"][axis] + (self["sw"][axis] / 2)
        min_ppm = self["o1p"][axis] - (self["sw"][axis] / 2)
        full_scale = np.linspace(max_ppm, min_ppm, int(self["si"][axis]))
        return full_scale[self._ppm_to_slice(axis, bounds)]

    def hz_scale(self,
                 axis: int
                 ) -> np.ndarray:
        # These use SFO, not BF
        max_hz = self["o1"][axis] + (self["sw"][axis] / (2 * self["sfo1"][axis]))
        min_hz = self["o1"][axis] - (self["sw"][axis] / (2 * self["sfo1"][axis]))
        return np.linspace(max_hz[axis], min_hz[axis], int(self["si"][axis]))

    def project(self,
                type: str,
                axis: Union[int, str],
                sign: str,
                bounds: TBounds = "",
                ) -> Dataset1DProjVirtual:
        if type not in ["projection", "sum"]:
            raise ValueError(f"Invalid value for type '{type}'")
        # Convenience for people who can't remember (like me)
        if axis == "column":  # sum / projection of columns
            axis = 0
        elif axis == "row":   # sum / projection of rows
            axis = 1
        if axis not in [0, 1]:
            raise ValueError(f"Invalid value for axis '{axis}'")
        # Allow some short forms...
        if sign not in ["positive", "negative", "pos", "neg"]:
            raise ValueError(f"Invalid value for sign '{sign}'")
        if sign == "pos":
            sign = "positive"
        elif sign == "neg":
            sign = "negative"
        # For some reason mypy doesn't realise that axis must be an int by here.
        index_bounds = self._ppm_to_slice(axis=(1 - axis), bounds=bounds)  # type: ignore
        return Dataset1DProjVirtual(self.path, proj_type=type,
                                    proj_axis=axis, sign=sign,
                                    index_bounds=index_bounds)

    def slice(self,
              axis: Union[int, str],
              ppm: float,
              ) -> Dataset1DProjVirtual:
        # convenience for people who can't remember (like me)
        if axis == "column":  # extract a column
            axis = 0
        elif axis == "row":   # extract a row
            axis = 1
        if axis not in [0, 1]:
            raise ValueError(f"Invalid value for axis '{axis}'")
        # For some reason mypy doesn't realise that axis must be an int by here.
        index = self._ppm_to_index(axis=(1 - axis), ppm=ppm)   # type: ignore
        return Dataset1DProjVirtual(self.path, proj_type="slice",
                                    proj_axis=axis, sign=None, index=index)


class Dataset1DProjVirtual(Dataset1DProj):
    """
    Class for a 1D projection generated from a 2D dataset.
    Note that the projection won't have any baseline correction, therefore
    integrals may not be entirely accurate.
    """

    def __init__(self,
                 path: Union[str, Path],
                 **kwargs
                 ) -> None:
        # Set some flags so that our overriding _read_spec() method can
        # calculate the projection
        self.proj_type = kwargs["proj_type"]  # "sum", "projection", or "slice"
        self.proj_axis = kwargs["proj_axis"]  # 0 or 1
        self.sign = kwargs.get("sign", None)  # "positive" or "negative"
        self.index_bounds = kwargs.get("index_bounds", None)  # only for sum or projection
        self.index = kwargs.get("index", None) # only for slice
        # Carry out the same initialisation tasks.
        # This calls _initialise_pars.
        Dataset1DProj.__init__(self, path)
        # Pars have been initialised at this point, so we can overwrite SI.
        # For 2D spectra SI is an array, for a standard projection it's a
        # single number. Not sure what other parameters need to be overwritten.
        self["si"] = self["si"][self.proj_axis]

    def _initialise_pars(self):
        # We don't want to inherit this from Dataset1DProj, because that
        # tries to look for the used_from file, and we don't have that.
        _Dataset._initialise_pars(self)

    def _read_spec(self) -> np.ndarray:
        # There's no 1r or 1i to read here, since the data isn't
        # from TopSpin.
        del self._p_real
        # First get the real, real part.
        _p_rr = self.path / "2rr"
        rr = np.fromfile(_p_rr, dtype=np.int32)
        rr = rr.reshape(int(self["si"][0]), int(self["si"][1]))
        rr = rr * (2 ** self["nc_proc"][1])
        # First check if it's a slice; that's the easiest case
        if self.proj_type == "slice":
            if self.proj_axis == 0:  # a column
                self.real = rr[:, self.index]
            elif self.proj_axis == 1:  # a row
                self.real = rr[self.index, :]
            return
        # Then check if there are bounds
        if self.index_bounds is not None:
            if self.proj_axis == 0:  # columns
                rr = rr[:, self.index_bounds]
            elif self.proj_axis == 1:  # rows
                rr = rr[self.index_bounds, :]
        # Zero all entries that are of the wrong sign
        if self.sign == "positive":
            rr[rr < 0] = 0
            projection_fn = np.amax
        elif self.sign == "negative":
            rr[rr > 0] = 0
            projection_fn = np.amin
        # Then make the projection / sum
        if self.proj_type == "projection":
            self.real = projection_fn(rr, axis=(1 - self.proj_axis))
        elif self.proj_type == "sum":
            self.real = np.sum(rr, axis=(1 - self.proj_axis))


TDataset1D = Union[Dataset1D, Dataset1DProj, Dataset1DProjVirtual]
TDatasetnD = Union[Dataset1D, Dataset1DProj, Dataset2D, Dataset1DProjVirtual]
